import json
import pathlib
import typing

import pandas
import pydantic
import pynwb
import typing_extensions


class Channel(pydantic.BaseModel):
    channel_id: str
    electrode_id: str
    type: typing.Literal["EXT"] = "EXT"  # TODO
    unit: typing.Literal["V"] = "V"
    sampling_frequency: float
    gain: float

    model_config = pydantic.ConfigDict(
        validate_assignment=True,  # Re-validate model on mutation
        extra="allow",  # Allow additional custom fields
    )


class ChannelTable(pydantic.BaseModel):
    channels: list[Channel]

    @classmethod
    @pydantic.validate_call
    def from_nwbfiles(cls, nwbfiles: list[pydantic.InstanceOf[pynwb.NWBFile]]) -> typing_extensions.Self | None:
        if len(nwbfiles) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)
        nwbfile = nwbfiles[0]

        electrical_series = [
            neurodata_object
            for neurodata_object in nwbfile.objects.values()
            if isinstance(neurodata_object, pynwb.ecephys.ElectricalSeries)
        ]
        if any(electrical_series) is False:
            return None

        channels = [
            Channel(
                channel_id=(
                    str(channel_name.values[0])
                    if (channel_name := electrode.get("channel_name", None)) is not None
                    else str(electrode.index[0])
                ),
                electrode_id=(
                    str(contact_ids.values[0])
                    if (contact_ids := electrode.get("contact_ids", None)) is not None
                    else str(electrode.index[0])
                ),
                type="EXT",
                unit="V",
                sampling_frequency=electrical_series[0].rate,  # TODO: generalize
                gain=electrical_series[0].conversion,
            )
            for electrode in nwbfile.electrodes
        ]
        return cls(channels=channels)

    @pydantic.validate_call
    def to_tsv(self, file_path: str | pathlib.Path):
        """
        Write the channels data to a TSV file.

        Parameters
        ----------
        file_path : path
            The path where the TSV file will be saved.
        """
        data_frame = pandas.DataFrame(data=[channel.model_dump() for channel in self.channels])
        data_frame.to_csv(path_or_buf=file_path, sep="\t", index=False)

    @pydantic.validate_call
    def to_json(self, file_path: str | pathlib.Path) -> None:
        """
        Save the channels information to a JSON file.

        Parameters
        ----------
        file_path : path
            The path to the output JSON file.
        """
        with file_path.open(mode="w") as file_stream:
            json.dump(obj=dict(), fp=file_stream, indent=4)



import pathlib
import shutil
import typing

import pydantic
import pynwb
import typing_extensions

from nwb2bids._converters._handle_file_mode import _handle_file_mode
from nwb2bids.bids_models import BidsSessionMetadata


class SessionConverter(pydantic.BaseModel):
    """
    Initialize a converter of NWB files to BIDS format.

    Parameters
    ----------
    nwbfile_paths : list of file paths
        The list of paths to NWB files associated with this session.
    """

    session_id: str = pydantic.Field(
        description="A unique session identifier.",
        pattern=r"^[^_]+$",  # No underscores allowed
    )
    nwbfile_paths: list[pydantic.FilePath] = pydantic.Field(
        description="List of NWB file paths which share this session ID.", min_length=1
    )
    session_metadata: BidsSessionMetadata | None = pydantic.Field(
        description="BIDS metadata extracted for this session.", default=None
    )

    @classmethod
    @pydantic.validate_call
    def from_nwb_directory(cls, nwb_directory: pydantic.DirectoryPath) -> list[typing_extensions.Self]:
        """
        Initialize a list of session converters from a list of NWB file paths.

        Automatically associates multiple NWB files to a single session according to session ID.

        Parameters
        ----------
        nwb_directory : directory path
            The path to the directory containing NWB files.

        Returns
        -------
        A list of SessionConverter instances, one per unique session ID.
        """
        nwb_file_paths = list(nwb_directory.rglob(pattern="*.nwb"))
        nwb_file_path_to_session_id = {
            nwb_file_path: pynwb.read_nwb(nwb_file_path).session_id for nwb_file_path in nwb_file_paths
        }  # IDEA: if this is too slow, could do direct h5py read instead, to avoid reading the entire file metadata

        unique_session_ids = set(nwb_file_path_to_session_id.values())
        unique_session_id_to_nwb_file_paths = {
            unique_session_id: [
                nwb_file_path
                for nwb_file_path, session_id in nwb_file_path_to_session_id.items()
                if session_id == unique_session_id
            ]
            for unique_session_id in unique_session_ids
        }

        session_converters = [
            cls(session_id=session_id, nwbfile_paths=nwb_file_paths)
            for session_id, nwb_file_paths in unique_session_id_to_nwb_file_paths.items()
        ]
        return session_converters

    def extract_session_metadata(self) -> None:
        """
        Extract metadata from the NWB file across the specified directory.
        """
        self.session_metadata = BidsSessionMetadata.from_nwbfile_paths(nwbfile_paths=self.nwbfile_paths)

    @pydantic.validate_call
    def convert_to_bids_session(
        self, bids_directory: str | pathlib.Path, file_mode: typing.Literal["move", "copy", "symlink"] | None = None
    ) -> None:
        """
        Convert the NWB file to a BIDS session directory.

        Parameters
        ----------
        bids_directory : directory path
            The path to the directory where the BIDS dataset will be created.
        file_mode : one of "move", "copy", or "symlink"
            Specifies how to handle the NWB files when converting to BIDS format.
              - "move": Move the files to the BIDS directory.
              - "copy": Copy the files to the BIDS directory.
              - "symlink": Create symbolic links to the files in the BIDS directory.
            The default behavior is to attempt to use symlinks, but fall back to copying if symlinks are not supported.
        """
        if len(self.nwbfile_paths) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)
        file_mode = _handle_file_mode(file_mode=file_mode)

        if self.session_metadata is None:
            self.extract_session_metadata()

        subject_id = self.session_metadata.participant.participant_id
        session_subdirectory = bids_directory / f"sub-{subject_id}" / f"ses-{self.session_id}" / "ecephys"
        session_subdirectory.mkdir(parents=True, exist_ok=True)

        self.write_ecephys_files(bids_directory=bids_directory)
        if self.session_metadata.events is not None:
            self.write_events_files(bids_directory=bids_directory)

        # TODO: determine icephys or ecephys suffix
        for nwbfile_path in self.nwbfile_paths:
            session_file_path = session_subdirectory / f"sub-{subject_id}_ses-{self.session_id}_ecephys.nwb"
            if file_mode == "copy":
                shutil.copy(src=nwbfile_path, dst=session_file_path)
            elif file_mode == "move":
                shutil.move(src=nwbfile_path, dst=session_file_path)
            elif file_mode == "symlink":
                session_file_path.symlink_to(target=nwbfile_path)

    @pydantic.validate_call
    def write_ecephys_files(self, bids_directory: str | pathlib.Path) -> None:
        """
        Write the `_probes`, `_channels`, and `_electrodes` metadata files, both `.tsv` and `.json`, for this session .

        Parameters
        ----------
        bids_directory : directory path
            The path to the directory where the BIDS dataset will be created.
        """
        if len(self.nwbfile_paths) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)

        if (
            self.session_metadata.probe_table is None
            and self.session_metadata.channel_table is None
            and self.session_metadata.electrode_table is None
        ):
            return

        self._establish_bids_directory_and_check_metadata(bids_directory=bids_directory)

        ecephys_directory = self._establish_ecephys_subdirectory(bids_directory=bids_directory)
        file_prefix = f"sub-{self.session_metadata.participant.participant_id}_ses-{self.session_id}"

        if self.session_metadata.probe_table is not None:
            probes_tsv_file_path = ecephys_directory / f"{file_prefix}_probes.tsv"
            self.session_metadata.probe_table.to_tsv(file_path=probes_tsv_file_path)

            probes_json_file_path = ecephys_directory / f"{file_prefix}_probes.json"
            self.session_metadata.probe_table.to_json(file_path=probes_json_file_path)

        if self.session_metadata.channel_table is not None:
            channels_tsv_file_path = ecephys_directory / f"{file_prefix}_channels.tsv"
            self.session_metadata.channel_table.to_tsv(file_path=channels_tsv_file_path)

            channels_json_file_path = ecephys_directory / f"{file_prefix}_channels.json"
            self.session_metadata.channel_table.to_json(file_path=channels_json_file_path)

        if self.session_metadata.electrode_table is not None:
            electrodes_tsv_file_path = ecephys_directory / f"{file_prefix}_electrodes.tsv"
            self.session_metadata.electrode_table.to_tsv(file_path=electrodes_tsv_file_path)

            electrodes_json_file_path = ecephys_directory / f"{file_prefix}_electrodes.json"
            self.session_metadata.electrode_table.to_json(file_path=electrodes_json_file_path)

    @pydantic.validate_call
    def write_events_files(self, bids_directory: str | pathlib.Path) -> None:
        """
        Write the `_events.tsv` and `_events.json` files for this session.

        Parameters
        ----------
        bids_directory : directory path
            The path to the directory where the BIDS dataset will be created.
        """
        if self.session_metadata.events is None:
            message = "No events metadata found in the session metadata - unable to write events TSV."
            raise ValueError(message)
        if len(self.nwbfile_paths) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)
        self._establish_bids_directory_and_check_metadata(bids_directory=bids_directory)

        ecephys_directory = self._establish_ecephys_subdirectory(bids_directory=bids_directory)
        file_prefix = f"sub-{self.session_metadata.participant.participant_id}_ses-{self.session_id}"
        session_events_tsv_file_path = ecephys_directory / f"{file_prefix}_events.tsv"
        self.session_metadata.events.to_tsv(file_path=session_events_tsv_file_path)

        # TODO: add merging from top-level additional metadata mechanism
        session_events_metadata_file_path = ecephys_directory / f"{file_prefix}_events.json"
        self.session_metadata.events.to_json(file_path=session_events_metadata_file_path)

        # IDEA: check events.json files, see if all are the same and if so remove the duplicates and move to outer level

    def _establish_bids_directory_and_check_metadata(self, bids_directory: pathlib.Path) -> None:
        bids_directory.mkdir(exist_ok=True)
        if self.session_metadata is None:
            self.extract_session_metadata()

    def _establish_ecephys_subdirectory(self, bids_directory: pathlib.Path) -> pathlib.Path:
        subject_directory = bids_directory / f"sub-{self.session_metadata.participant.participant_id}"
        subject_directory.mkdir(exist_ok=True)
        session_directory = subject_directory / f"ses-{self.session_id}"
        session_directory.mkdir(exist_ok=True)
        ecephys_directory = session_directory / "ecephys"
        ecephys_directory.mkdir(exist_ok=True)

        return ecephys_directory



"""Integration tests for the primary `convert_nwb_dataset` function."""

import pathlib

import nwb2bids


def test_minimal_convert_nwb_dataset(minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path):
    nwb2bids.convert_nwb_dataset(nwb_directory=minimal_nwbfile_path.parent, bids_directory=temporary_bids_directory)

    expected_structure = {
        temporary_bids_directory: {"directories": {"sub-123"}, "files": {"participants.json", "participants.tsv"}},
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


def test_ecephys_convert_nwb_dataset(ecephys_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path):
    nwb2bids.convert_nwb_dataset(nwb_directory=ecephys_nwbfile_path.parent, bids_directory=temporary_bids_directory)

    expected_structure = {
        temporary_bids_directory: {"directories": {"sub-123"}, "files": {"participants.json", "participants.tsv"}},
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
                "sub-123_ses-456_channels.tsv",
                "sub-123_ses-456_channels.json",
                "sub-123_ses-456_electrodes.tsv",
                "sub-123_ses-456_electrodes.json",
                "sub-123_ses-456_probes.tsv",
                "sub-123_ses-456_probes.json",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )



import typing
import warnings

import click

from .._base import convert_nwb_dataset


# nwb2bids
@click.group(name="nwb2bids")
def _nwb2bids_cli():
    pass


# TODO: remove
# nwb2bids reposit < nwb_directory > < bids_directory >
@_nwb2bids_cli.command(name="reposit", deprecated=True)
@click.argument("nwb_directory", type=click.Path(writable=False))
@click.argument("bids_directory", type=click.Path(writable=True))
@click.option(
    "--no-copy",
    help="Whether or not to copy data from the NWB files into the BIDS format. Otherwise, files will be moved.",
    required=False,
    type=click.BOOL,
    default=False,
)
@click.option(
    "--additional-metadata-file-path",
    "additional_metadata_file_path",
    help=(
        "Path to a JSON file containing additional metadata to be included in the BIDS dataset. "
        "This file should contain a dictionary with keys corresponding to BIDS entities."
    ),
    required=False,
    type=click.Path(exists=True, dir_okay=False, readable=True),
    default=None,
)
def _run_convert_nwb_dataset(
    nwb_directory: str,
    bids_directory: str,
    no_copy: bool = False,
    additional_metadata_file_path: str | None = None,
) -> None:
    message = (
        "The `nwb2bids reposit` CLI is deprecated and will be removed in a future release. "
        "Please use `nwb2bids convert` instead.",
    )
    warnings.warn(message=message, category=DeprecationWarning, stacklevel=2)

    convert_nwb_dataset(
        nwb_directory=nwb_directory,
        bids_directory=bids_directory,
        file_mode="move" if no_copy is True else "copy",
        additional_metadata_file_path=additional_metadata_file_path,
    )


# nwb2bids convert < nwb_directory > < bids_directory >
@_nwb2bids_cli.command(name="convert")
@click.argument("nwb_directory", type=click.Path(writable=False))
@click.argument("bids_directory", type=click.Path(writable=True))
@click.option(
    "--file-mode",
    help=(
        "How to handle the source NWB files when converting to BIDS format. "
        "By default, will attempt to utilize 'symlink' if the system allows. "
        "Otherwise, will 'copy' to preserve directory integrity. "
        "Use 'move' for fastest speeds if you do not wish to keep the original NWB directory structure."
    ),
    required=False,
    type=click.Choice(["copy", "move", "symlink"], case_sensitive=False),
    default=True,
)
@click.option(
    "--additional-metadata-file-path",
    "additional_metadata_file_path",
    help=(
        "Path to a JSON file containing additional metadata to be included in the BIDS dataset. "
        "This file should contain a dictionary with keys corresponding to BIDS entities."
    ),
    required=False,
    type=click.Path(exists=True, dir_okay=False, readable=True),
    default=None,
)
def _run_convert_nwb_dataset(
    nwb_directory: str,
    bids_directory: str,
    file_mode: typing.Literal["copy", "move", "symlink"] | None = None,
    additional_metadata_file_path: str | None = None,
) -> None:
    """
    Convert NWB files to BIDS format.

    NWB_DIRECTORY : The path to the folder containing NWB files.
    BIDS_DIRECTORY : The path to the folder where the BIDS dataset will be created.
    """
    convert_nwb_dataset(
        nwb_directory=nwb_directory,
        bids_directory=bids_directory,
        file_mode=file_mode,
        additional_metadata_file_path=additional_metadata_file_path,
    )



import pydantic
import pynwb
import typing_extensions

from ._channels import ChannelTable
from ._electrodes import ElectrodeTable
from ._events import Events
from ._participant import Participant
from ._probes import ProbeTable


class BidsSessionMetadata(pydantic.BaseModel):
    """
    Schema for the metadata of a single BIDS session.
    """

    session_id: str = pydantic.Field(
        description="A unique session identifier.",
        pattern=r"^[^_]+$",  # No underscores allowed
    )
    participant: Participant = pydantic.Field(description="Metadata about a participant used in this experiment.")
    # general_metadata: GeneralMetadata = pydantic.Field(
    #     description="General metadata about the setup for this experiment."
    # )
    events: Events | None = pydantic.Field(
        description="Timing data and metadata regarding events that occur during this experiment.", default=None
    )
    probe_table: ProbeTable | None = None
    channel_table: ChannelTable | None = None
    electrode_table: ElectrodeTable | None = None

    @classmethod
    @pydantic.validate_call
    def from_nwbfile_paths(cls, nwbfile_paths: list[pydantic.FilePath]) -> typing_extensions.Self:
        nwbfiles = [pynwb.read_nwb(nwbfile_path) for nwbfile_path in nwbfile_paths]
        session_ids = list({nwbfile.session_id for nwbfile in nwbfiles})
        if len(session_ids) > 1:
            message = "Multiple differing session IDs found - please check how this method was called."
            raise ValueError(message)
        session_id = session_ids[0]

        participant = Participant.from_nwbfiles(nwbfiles=nwbfiles)
        events = Events.from_nwbfiles(nwbfiles=nwbfiles)
        probe_table = ProbeTable.from_nwbfiles(nwbfiles=nwbfiles)
        channel_table = ChannelTable.from_nwbfiles(nwbfiles=nwbfiles)
        electrode_table = ElectrodeTable.from_nwbfiles(nwbfiles=nwbfiles)

        dictionary = {
            "session_id": session_id,
            "participant": participant,
        }
        if events is not None:
            dictionary["events"] = events
        if probe_table is not None:
            dictionary["probe_table"] = probe_table
        if channel_table is not None:
            dictionary["channel_table"] = channel_table
        if electrode_table is not None:
            dictionary["electrode_table"] = electrode_table

        session_metadata = BidsSessionMetadata(**dictionary)
        return session_metadata



import collections
import json
import pathlib
import typing

import pandas
import pydantic
import typing_extensions

from nwb2bids.bids_models import DatasetDescription

from ._session_converter import SessionConverter


class DatasetConverter(pydantic.BaseModel):
    session_converters: list[SessionConverter] = pydantic.Field(
        description="List of session converters. Typically instantiated by calling `.from_nwb_directory()`.",
        min_length=1,
    )
    dataset_description: DatasetDescription | None = pydantic.Field(
        description="The BIDS-compatible dataset description.",
        default=None,
    )

    @classmethod
    @pydantic.validate_call
    def from_nwb_directory(
        cls, nwb_directory: pydantic.DirectoryPath, additional_metadata_file_path: pydantic.FilePath | None = None
    ) -> typing_extensions.Self:
        """
        Initialize a converter of NWB files to BIDS format.

        Parameters
        ----------
        nwb_directory : directory path
            The path to the directory containing NWB files.
        additional_metadata_file_path : file path, optional
            The path to a JSON file containing additional metadata to be included in the BIDS dataset.
            If not provided, the method will also look for a file named "additional_metadata.json" in the NWB directory.
        """
        session_converters = SessionConverter.from_nwb_directory(nwb_directory=nwb_directory)

        dataset_description = None
        additional_metadata_file_path = (
            secondary_path
            if additional_metadata_file_path is None
            and (secondary_path := nwb_directory / "additional_metadata.json").exists()
            else additional_metadata_file_path
        )
        if additional_metadata_file_path is not None:
            dataset_description = DatasetDescription.from_file_path(file_path=additional_metadata_file_path)

        dataset_converter = cls(session_converters=session_converters, dataset_description=dataset_description)
        return dataset_converter

    def extract_dataset_metadata(self) -> None:
        """
        Deploy the call to `.extract_session_metadata()` for each session converter.
        """
        collections.deque(
            (
                session_converter.extract_session_metadata()
                for session_converter in self.session_converters
                if session_converter.session_metadata is None
            ),
            maxlen=0,
        )

    @pydantic.validate_call
    def convert_to_bids_dataset(
        self, bids_directory: str | pathlib.Path, file_mode: typing.Literal["move", "copy", "symlink"] | None = None
    ) -> None:
        """
        Convert the directory of NWB files to a BIDS dataset.

        Parameters
        ----------
        bids_directory : directory path
            The path to the directory where the BIDS dataset will be created.
        file_mode : one of "move", "copy", or "symlink"
            Specifies how to handle the NWB files when converting to BIDS format.
            - "move": Move the files to the BIDS directory.
            - "copy": Copy the files to the BIDS directory.
            - "symlink": Create symbolic links to the files in the BIDS directory.
        """
        self._establish_bids_directory_and_check_metadata(bids_directory=bids_directory)

        if self.dataset_description is not None:
            self.write_dataset_description(bids_directory=bids_directory)

        self.write_participants_metadata(bids_directory=bids_directory)
        self.write_sessions_metadata(bids_directory=bids_directory)

        collections.deque(
            (
                session_converter.convert_to_bids_session(bids_directory=bids_directory, file_mode=file_mode)
                for session_converter in self.session_converters
            ),
            maxlen=0,
        )

    @pydantic.validate_call
    def write_dataset_description(self, bids_directory: str | pathlib.Path) -> None:
        self._establish_bids_directory_and_check_metadata(bids_directory=bids_directory)

        dataset_description_dictionary = self.dataset_description.model_dump()

        dataset_description_file_path = bids_directory / "dataset_description.json"
        with dataset_description_file_path.open(mode="w") as file_stream:
            json.dump(obj=dataset_description_dictionary, fp=file_stream, indent=4)

    @pydantic.validate_call
    def write_participants_metadata(self, bids_directory: str | pathlib.Path) -> None:
        """
        Write the `participants.tsv` and `participants.json` files.

        Parameters
        ----------
        bids_directory : directory path
            The path to the directory where the BIDS dataset will be created.
        """
        self._establish_bids_directory_and_check_metadata(bids_directory=bids_directory)

        participants_data_frame = pandas.DataFrame.from_records(
            data=[
                {
                    key: value
                    for key, value in session_converter.session_metadata.participant.model_dump().items()
                    if value is not None
                }
                for session_converter in self.session_converters
            ]
        ).astype("string")

        # BIDS requires sub- prefix in table values
        participants_data_frame["participant_id"] = participants_data_frame["participant_id"].apply(
            lambda participant_id: f"sub-{participant_id}"
        )
        is_field_in_table = {field: True for field in participants_data_frame.keys()}

        # BIDS Validator is strict regarding column order
        required_column_order = [
            field
            for field in ["participant_id", "species", "sex", "strain"]
            if is_field_in_table.get(field, False) is True
        ]
        column_order = required_column_order + [
            field
            for field in participants_data_frame.columns
            if is_field_in_table.get(field, False) is True and field not in required_column_order
        ]

        participants_tsv_file_path = bids_directory / "participants.tsv"
        participants_data_frame.to_csv(
            path_or_buf=participants_tsv_file_path, mode="w", index=False, sep="\t", columns=column_order
        )

        is_field_in_table = {field: True for field in participants_data_frame.keys()}
        participants_schema = self.session_converters[0].session_metadata.participant.model_json_schema()
        participants_json = {
            field: info["description"]
            for field, info in participants_schema["properties"].items()
            if is_field_in_table.get(field, False) is True
        }
        participants_json_file_path = bids_directory / "participants.json"
        with participants_json_file_path.open(mode="w") as file_stream:
            json.dump(obj=participants_json, fp=file_stream, indent=4)

    @pydantic.validate_call
    def write_sessions_metadata(self, bids_directory: str | pathlib.Path) -> None:
        """
        Write the `_sessions.tsv` and `_sessions.json` files, then create empty participant and session directories.

        Parameters
        ----------
        bids_directory : directory path
            The path to the directory where the BIDS dataset will be created.
        """
        self._establish_bids_directory_and_check_metadata(bids_directory=bids_directory)

        subject_id_to_sessions = collections.defaultdict(list)
        for session_converter in self.session_converters:
            subject_id_to_sessions[session_converter.session_metadata.participant.participant_id].append(
                session_converter
            )

        # TODO: expand beyond just session_id field (mainly via additional metadata)
        sessions_schema = self.session_converters[0].session_metadata.model_json_schema()
        sessions_json = {"session_id": sessions_schema["properties"]["session_id"]["description"]}

        for subject_id, sessions_metadata in subject_id_to_sessions.items():
            subject_directory = bids_directory / f"sub-{subject_id}"
            subject_directory.mkdir(exist_ok=True)

            # BIDS requires ses- prefix in table values
            sessions_data_frame = pandas.DataFrame(
                {"session_id": [f"ses-{session_metadata.session_id}" for session_metadata in sessions_metadata]}
            )

            session_tsv_file_path = subject_directory / f"sub-{subject_id}_sessions.tsv"
            sessions_data_frame.to_csv(path_or_buf=session_tsv_file_path, mode="w", index=False, sep="\t")

            session_json_file_path = subject_directory / f"sub-{subject_id}_sessions.json"
            with session_json_file_path.open(mode="w") as file_stream:
                json.dump(obj=sessions_json, fp=file_stream, indent=4)

            for session_metadata in sessions_metadata:
                session_directory = subject_directory / f"ses-{session_metadata.session_id}"
                session_directory.mkdir(exist_ok=True)

    def _establish_bids_directory_and_check_metadata(self, bids_directory: pathlib.Path) -> None:
        bids_directory.mkdir(exist_ok=True)
        if any(session_converter.session_metadata is None for session_converter in self.session_converters):
            self.extract_dataset_metadata()



import pathlib
import shutil
import tempfile
import typing


def _handle_file_mode(
    file_mode: typing.Literal["move", "copy", "symlink"] | None = None,
) -> typing.Literal["move", "copy", "symlink"]:
    if file_mode is not None:
        return file_mode

    try:
        test_directory = pathlib.Path(tempfile.mkdtemp(prefix="nwb2bids-"))
        test_file_path = test_directory / "test_file.txt"
        test_file_path.touch()
        (test_directory / "test_symlink.txt").symlink_to(target=test_file_path)
        shutil.rmtree(path=test_directory, ignore_errors=True)
        return "symlink"
    except (OSError, PermissionError):  # Windows can sometimes have trouble with symlinks
        return "copy"



import warnings

from ._convert_nwb_dataset import convert_nwb_dataset


def reposit(
    in_dir: str, out_dir: str, *, no_copy: bool = False, additional_metadata_file_path: str | None = None
) -> None:
    message = (
        "The `nwb2bids.reposit` function is deprecated and will be removed in a future release. "
        "Please use `nwb2bids.convert_nwb_dataset` instead.",
    )
    warnings.warn(message=message, category=DeprecationWarning, stacklevel=2)

    convert_nwb_dataset(
        nwb_directory=in_dir,
        out_dir=out_dir,
        file_mode="move" if no_copy is True else "copy",
        additional_metadata_file_path=additional_metadata_file_path,
    )



import typing

import pydantic
import pynwb
import typing_extensions


class Participant(pydantic.BaseModel):
    participant_id: str = pydantic.Field(
        description="A unique identifier for this participant.",
        pattern=r"^[^_]+$",  # No underscores allowed
    )
    species: str = pydantic.Field(
        description=(
            "The species should be the proper Latin binomial species name from the NCBI Taxonomy "
            "(for example, Mus musculus)."
        ),
        pattern=r"([A-Z][a-z]* [a-z]+)|(http://purl.obolibrary.org/obo/NCBITaxon_\d+)",  # Latin binomial or NCBI link
        # TODO: see if BIDS validator accepts purl.obolib links directly
    )
    sex: typing.Literal[
        "male", "m", "M", "MALE", "Male", "female", "f", "F", "FEMALE", "Female", "other", "o", "O", "OTHER", "Other"
    ] = pydantic.Field(
        description=(
            'String value indicating phenotypical sex, one of "male", "female", "other".\n'
            '\tFor "male", use one of these values: male, m, M, MALE, Male.\n'
            '\tFor "female", use one of these values: female, f, F, FEMALE, Female.\n'
            '\tFor "other", use one of these values: other, o, O, OTHER, Other.'
        )
    )
    strain: str | None = pydantic.Field(
        description=(
            "For species other than Homo sapiens, the string value indicating the strain of the species "
            "(for example, C57BL/6J)."
        ),
        default=None,
    )

    model_config = pydantic.ConfigDict(
        validate_assignment=True,  # Re-validate model on mutation
        extra="allow",  # Allow additional custom fields
    )

    @classmethod
    @pydantic.validate_call
    def from_nwbfiles(cls, nwbfiles: list[pydantic.InstanceOf[pynwb.NWBFile]]) -> typing_extensions.Self:
        """
        Extracts participant metadata from the in-memory NWBFile objects.
        """
        if len(nwbfiles) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)

        nwbfile = nwbfiles[0]

        subject = cls(
            participant_id=nwbfile.subject.subject_id.replace("_", "-"),
            species=nwbfile.subject.species,
            strain=nwbfile.subject.strain,
            sex=nwbfile.subject.sex,
            # TODO: add more
            # birthday=nwbfile.participant.date_of_birth,
            # age=nwbfile.participant.age
        )
        return subject



from ._bids_microephys_session_metadata import BidsSessionMetadata, Participant
from ._dataset_description import DatasetDescription

__all__ = [
    "BidsSessionMetadata",
    "DatasetDescription",
    "Participant",
]



import pathlib
import typing

import pydantic

from nwb2bids._converters._dataset_converter import DatasetConverter


@pydantic.validate_call
def convert_nwb_dataset(
    *,
    nwb_directory: pydantic.DirectoryPath,
    bids_directory: str | pathlib.Path,
    file_mode: typing.Literal["move", "copy", "symlink"] | None = None,
    additional_metadata_file_path: pydantic.FilePath | None = None,
) -> None:
    converter = DatasetConverter.from_nwb_directory(
        nwb_directory=nwb_directory, additional_metadata_file_path=additional_metadata_file_path
    )
    converter.extract_dataset_metadata()
    converter.convert_to_bids_dataset(bids_directory=bids_directory, file_mode=file_mode)
